# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------


def cf_databoxedge_cl(cli_ctx, *_):
    from azure.cli.core.commands.client_factory import get_mgmt_service_client
    from azure.cli.core.profiles import ResourceType
    return get_mgmt_service_client(cli_ctx,
                                   ResourceType.MGMT_DATABOXEDGE)


def cf_device(cli_ctx, *_):
    return cf_databoxedge_cl(cli_ctx).devices


def cf_alert(cli_ctx, *_):
    return cf_databoxedge_cl(cli_ctx).alerts


def cf_bandwidth_schedule(cli_ctx, *_):
    return cf_databoxedge_cl(cli_ctx).bandwidth_schedules


def cf_job(cli_ctx, *_):
    return cf_databoxedge_cl(cli_ctx).jobs


def cf_node(cli_ctx, *_):
    return cf_databoxedge_cl(cli_ctx).nodes


def cf_order(cli_ctx, *_):
    return cf_databoxedge_cl(cli_ctx).orders


def cf_sku(cli_ctx, *_):
    return cf_databoxedge_cl(cli_ctx).skus
